"use strict";
/**
 * Copyright 2019 Google LLC. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const chalk_1 = require("chalk");
const semver = require("semver");
const stream_1 = require("stream");
const checkpoint_1 = require("./util/checkpoint");
const concat = require('concat-stream');
const conventionalCommitsFilter = require('conventional-commits-filter');
const conventionalCommitsParser = require('conventional-commits-parser');
const conventionalChangelogWriter = require('conventional-changelog-writer');
const parseGithubRepoUrl = require('parse-github-repo-url');
const presetFactory = require('conventional-changelog-conventionalcommits');
class ConventionalCommits {
    constructor(options) {
        const parsedGithubRepoUrl = parseGithubRepoUrl(options.githubRepoUrl);
        if (!parsedGithubRepoUrl)
            throw Error('could not parse githubRepoUrl');
        const [owner, repository] = parsedGithubRepoUrl;
        this.commits = options.commits;
        this.bumpMinorPreMajor = options.bumpMinorPreMajor || false;
        this.host = options.host || 'https://www.github.com';
        this.owner = owner;
        this.repository = repository;
    }
    async suggestBump(version) {
        const preMajor = this.bumpMinorPreMajor
            ? semver.lt(version, 'v1.0.0')
            : false;
        const bump = await this.guessReleaseType(preMajor);
        checkpoint_1.checkpoint(`release as ${chalk_1.default.green(bump.releaseType)}: ${chalk_1.default.yellow(bump.reason)}`, checkpoint_1.CheckpointType.Success);
        return bump;
    }
    async generateChangelogEntry(options) {
        const context = {
            host: this.host,
            owner: this.owner,
            repository: this.repository,
            version: options.version,
            previousTag: options.previousTag,
            currentTag: options.currentTag,
            linkCompare: !!options.previousTag,
        };
        const preset = await presetFactory({});
        return new Promise((resolve, reject) => {
            let content = '';
            const stream = this.commitsReadable()
                .pipe(conventionalCommitsParser(preset.parserOpts))
                .pipe(conventionalChangelogWriter(context, preset.writerOpts));
            stream.on('error', (err) => {
                return reject(err);
            });
            stream.on('data', (buffer) => {
                content += buffer.toString('utf8');
            });
            stream.on('end', () => {
                return resolve(content.trim());
            });
        });
    }
    async guessReleaseType(preMajor) {
        const VERSIONS = ['major', 'minor', 'patch'];
        const preset = await presetFactory({ preMajor });
        return new Promise((resolve, reject) => {
            const stream = this.commitsReadable()
                .pipe(conventionalCommitsParser(preset.parserOpts))
                .pipe(concat((data) => {
                const commits = conventionalCommitsFilter(data);
                let result = preset.recommendedBumpOpts.whatBump(commits, preset.recommendedBumpOpts);
                if (result && result.level != null) {
                    result.releaseType = VERSIONS[result.level];
                }
                else if (result == null) {
                    result = {};
                }
                return resolve(result);
            }));
            stream.on('error', (err) => {
                return reject(err);
            });
        });
    }
    commitsReadable() {
        // The conventional commits parser expects an array of string commit
        // messages terminated by `-hash-` followed by the commit sha. We
        // piggyback off of this, and use this sha when choosing a
        // point to branch from for PRs.
        const commitsReadable = new stream_1.Readable();
        this.commits.forEach((commit) => {
            commitsReadable.push(`${commit.message}\n-hash-\n${commit.sha}`);
        });
        commitsReadable.push(null);
        return commitsReadable;
    }
}
exports.ConventionalCommits = ConventionalCommits;
//# sourceMappingURL=conventional-commits.js.map