"use strict";
/**
 * Copyright 2019 Google LLC. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const semver = require("semver");
const checkpoint_1 = require("./util/checkpoint");
const conventional_commits_1 = require("./conventional-commits");
const github_1 = require("./github");
const commit_split_1 = require("./commit-split");
// Generic
const changelog_1 = require("./updaters/changelog");
// JavaScript
const package_json_1 = require("./updaters/package-json");
const samples_package_json_1 = require("./updaters/samples-package-json");
// Yoshi PHP Monorepo
const php_client_version_1 = require("./updaters/php-client-version");
const php_manifest_1 = require("./updaters/php-manifest");
const root_composer_1 = require("./updaters/root-composer");
const version_1 = require("./updaters/version");
// Java
const pom_xml_1 = require("./updaters/pom-xml");
// Yoshi Java Auth Library
const java_auth_versions_1 = require("./updaters/java-auth-versions");
const java_auth_readme_1 = require("./updaters/java-auth-readme");
const parseGithubRepoUrl = require('parse-github-repo-url');
var ReleaseType;
(function (ReleaseType) {
    ReleaseType["Node"] = "node";
    ReleaseType["PHPYoshi"] = "php-yoshi";
    ReleaseType["JavaAuthYoshi"] = "java-auth-yoshi";
})(ReleaseType = exports.ReleaseType || (exports.ReleaseType = {}));
class ReleasePR {
    constructor(options) {
        this.bumpMinorPreMajor = options.bumpMinorPreMajor || false;
        this.labels = options.label.split(',');
        this.repoUrl = options.repoUrl;
        this.token = options.token;
        this.packageName = options.packageName;
        this.releaseAs = options.releaseAs;
        this.releaseType = options.releaseType;
        this.apiUrl = options.apiUrl;
        this.proxyKey = options.proxyKey;
        this.snapshot = options.snapshot;
        this.gh = this.gitHubInstance();
    }
    async run() {
        const pr = await this.gh.findMergedReleasePR(this.labels);
        if (pr) {
            // a PR already exists in the autorelease: pending state.
            checkpoint_1.checkpoint(`pull #${pr.number} ${pr.sha} has not yet been released`, checkpoint_1.CheckpointType.Failure);
        }
        else {
            switch (this.releaseType) {
                case ReleaseType.Node:
                    return this.nodeRelease();
                case ReleaseType.PHPYoshi:
                    return this.phpYoshiRelease();
                case ReleaseType.JavaAuthYoshi:
                    return this.javaAuthYoshiRelease();
                default:
                    throw Error('unknown release type');
            }
        }
    }
    async nodeRelease() {
        const latestTag = await this.gh.latestTag();
        const commits = await this.commits(latestTag ? latestTag.sha : undefined);
        const cc = new conventional_commits_1.ConventionalCommits({
            commits,
            githubRepoUrl: this.repoUrl,
            bumpMinorPreMajor: this.bumpMinorPreMajor,
        });
        const candidate = await this.coerceReleaseCandidate(cc, latestTag);
        const changelogEntry = await cc.generateChangelogEntry({
            version: candidate.version,
            currentTag: `v${candidate.version}`,
            previousTag: candidate.previousTag,
        });
        // don't create a release candidate until user facing changes
        // (fix, feat, BREAKING CHANGE) have been made; a CHANGELOG that's
        // one line is a good indicator that there were no interesting commits.
        if (changelogEmpty(changelogEntry)) {
            checkpoint_1.checkpoint(`no user facing commits found since ${latestTag ? latestTag.sha : 'beginning of time'}`, checkpoint_1.CheckpointType.Failure);
            return;
        }
        const updates = [];
        updates.push(new changelog_1.Changelog({
            path: 'CHANGELOG.md',
            changelogEntry,
            version: candidate.version,
            packageName: this.packageName,
        }));
        updates.push(new package_json_1.PackageJson({
            path: 'package.json',
            changelogEntry,
            version: candidate.version,
            packageName: this.packageName,
        }));
        updates.push(new samples_package_json_1.SamplesPackageJson({
            path: 'samples/package.json',
            changelogEntry,
            version: candidate.version,
            packageName: this.packageName,
        }));
        await this.openPR(commits[0].sha, `${changelogEntry}\n---\n`, updates, candidate.version);
    }
    async phpYoshiRelease() {
        const latestTag = await this.gh.latestTag();
        const commits = await this.commits(latestTag ? latestTag.sha : undefined);
        // we create an instance of conventional CHANGELOG for bumping the
        // top-level tag version we maintain on the mono-repo itself.
        const ccb = new conventional_commits_1.ConventionalCommits({
            commits,
            githubRepoUrl: this.repoUrl,
            bumpMinorPreMajor: true,
        });
        const candidate = await this.coerceReleaseCandidate(ccb, latestTag);
        // partition a set of packages in the mono-repo that need to be
        // updated since our last release -- the set of string keys
        // is sorted to ensure consistency in the CHANGELOG.
        const updates = [];
        let changelogEntry = `## ${candidate.version}`;
        const bulkUpdate = await this.releaseAllPHPLibraries(commits, updates, changelogEntry);
        changelogEntry = bulkUpdate.changelogEntry;
        // update the aggregate package information in the root
        // composer.json and manifest.json.
        updates.push(new root_composer_1.RootComposer({
            path: 'composer.json',
            changelogEntry,
            version: candidate.version,
            versions: bulkUpdate.versionUpdates,
            packageName: this.packageName,
        }));
        updates.push(new php_manifest_1.PHPManifest({
            path: 'docs/manifest.json',
            changelogEntry,
            version: candidate.version,
            versions: bulkUpdate.versionUpdates,
            packageName: this.packageName,
        }));
        updates.push(new changelog_1.Changelog({
            path: 'CHANGELOG.md',
            changelogEntry,
            version: candidate.version,
            packageName: this.packageName,
        }));
        ['src/Version.php', 'src/ServiceBuilder.php'].forEach((path) => {
            updates.push(new php_client_version_1.PHPClientVersion({
                path,
                changelogEntry,
                version: candidate.version,
                packageName: this.packageName,
            }));
        });
        await this.openPR(commits[0].sha, changelogEntry, updates, candidate.version);
    }
    async releaseAllPHPLibraries(commits, updates, changelogEntry) {
        const cs = new commit_split_1.CommitSplit();
        const commitLookup = cs.split(commits);
        const pkgKeys = Object.keys(commitLookup).sort();
        // map of library names that need to be updated in the top level
        // composer.json and manifest.json.
        const versionUpdates = {};
        // walk each individual library updating the VERSION file, and
        // if necessary the `const VERSION` in the client library.
        for (let i = 0; i < pkgKeys.length; i++) {
            const pkgKey = pkgKeys[i];
            const cc = new conventional_commits_1.ConventionalCommits({
                commits: commitLookup[pkgKey],
                githubRepoUrl: this.repoUrl,
                bumpMinorPreMajor: this.bumpMinorPreMajor,
            });
            // some packages in the mono-repo might have only had chores,
            // build updates, etc., applied.
            if (!changelogEmpty(await cc.generateChangelogEntry({ version: '0.0.0' }))) {
                try {
                    const contents = await this.gh.getFileContents(`${pkgKey}/VERSION`);
                    const bump = await cc.suggestBump(contents.parsedContent);
                    const candidate = semver.inc(contents.parsedContent, bump.releaseType);
                    if (!candidate) {
                        checkpoint_1.checkpoint(`failed to update ${pkgKey} version`, checkpoint_1.CheckpointType.Failure);
                        continue;
                    }
                    const meta = JSON.parse((await this.gh.getFileContents(`${pkgKey}/composer.json`))
                        .parsedContent);
                    versionUpdates[meta.name] = candidate;
                    changelogEntry = updatePHPChangelogEntry(`${meta.name} ${candidate}`, changelogEntry, await cc.generateChangelogEntry({ version: candidate }));
                    updates.push(new version_1.Version({
                        path: `${pkgKey}/VERSION`,
                        changelogEntry,
                        version: candidate,
                        packageName: this.packageName,
                        contents,
                    }));
                    // extra.component indicates an entry-point class file
                    // that must have its version # updatd.
                    if (meta.extra &&
                        meta.extra.component &&
                        meta.extra.component.entry) {
                        updates.push(new php_client_version_1.PHPClientVersion({
                            path: `${pkgKey}/${meta.extra.component.entry}`,
                            changelogEntry,
                            version: candidate,
                            packageName: this.packageName,
                        }));
                    }
                }
                catch (err) {
                    if (err.status === 404) {
                        // if the updated path has no VERSION, assume this isn't a
                        // module that needs updating.
                        continue;
                    }
                    else {
                        throw err;
                    }
                }
            }
        }
        return { changelogEntry, versionUpdates };
    }
    async javaAuthYoshiRelease() {
        const latestTag = await this.gh.latestTag();
        const commits = this.snapshot
            ? [
                {
                    sha: 'abc123',
                    message: 'fix: ',
                    files: [],
                },
            ]
            : await this.commits(latestTag ? latestTag.sha : undefined, 100, true);
        let prSHA = commits[0].sha;
        const cc = new conventional_commits_1.ConventionalCommits({
            commits,
            githubRepoUrl: this.repoUrl,
            bumpMinorPreMajor: this.bumpMinorPreMajor,
        });
        const candidate = await this.coerceReleaseCandidate(cc, latestTag);
        let changelogEntry = await cc.generateChangelogEntry({
            version: candidate.version,
            currentTag: `v${candidate.version}`,
            previousTag: candidate.previousTag,
        });
        // snapshot entries are special:
        // 1. they don't update the README or CHANGELOG.
        // 2. they always update a patch with the -SNAPSHOT suffix.
        // 3. they're haunted.
        if (this.snapshot) {
            prSHA = latestTag.sha;
            candidate.version = `${candidate.version}-SNAPSHOT`;
            changelogEntry =
                '### Updating meta-information for bleeding-edge SNAPSHOT release.';
        }
        // don't create a release candidate until user facing changes
        // (fix, feat, BREAKING CHANGE) have been made; a CHANGELOG that's
        // one line is a good indicator that there were no interesting commits.
        if (changelogEmpty(changelogEntry) && !this.snapshot) {
            checkpoint_1.checkpoint(`no user facing commits found since ${latestTag ? latestTag.sha : 'beginning of time'}`, checkpoint_1.CheckpointType.Failure);
            return;
        }
        const updates = [];
        if (!this.snapshot) {
            updates.push(new changelog_1.Changelog({
                path: 'CHANGELOG.md',
                changelogEntry,
                version: candidate.version,
                packageName: this.packageName,
            }));
            updates.push(new java_auth_readme_1.JavaAuthReadme({
                path: 'README.md',
                changelogEntry,
                version: candidate.version,
                packageName: this.packageName,
            }));
        }
        updates.push(new java_auth_versions_1.JavaAuthVersions({
            path: 'versions.txt',
            changelogEntry,
            version: candidate.version,
            packageName: this.packageName,
        }));
        [
            'appengine/pom.xml',
            'bom/pom.xml',
            'credentials/pom.xml',
            'oauth2_http/pom.xml',
            'pom.xml',
        ].forEach(path => {
            updates.push(new pom_xml_1.PomXML({
                path,
                changelogEntry,
                version: candidate.version,
                packageName: this.packageName,
            }));
        });
        await this.openPR(prSHA, `${changelogEntry}\n---\n`, updates, candidate.version);
    }
    async closeStaleReleasePRs(currentPRNumber) {
        const prs = await this.gh.findOpenReleasePRs(this.labels);
        for (let i = 0, pr; i < prs.length; i++) {
            pr = prs[i];
            // don't close the most up-to-date release PR.
            if (pr.number !== currentPRNumber) {
                checkpoint_1.checkpoint(`closing pull #${pr.number}`, checkpoint_1.CheckpointType.Failure);
                await this.gh.closePR(pr.number);
            }
        }
    }
    async coerceReleaseCandidate(cc, latestTag) {
        const previousTag = latestTag ? latestTag.name : undefined;
        let version = latestTag ? latestTag.version : '1.0.0';
        if (latestTag && !this.releaseAs) {
            const bump = await cc.suggestBump(version);
            const candidate = semver.inc(version, bump.releaseType);
            if (!candidate)
                throw Error(`failed to increment ${version}`);
            version = candidate;
        }
        else if (this.releaseAs) {
            version = this.releaseAs;
        }
        return { version, previousTag };
    }
    async commits(sha, perPage = 100, labels = false) {
        const commits = await this.gh.commitsSinceSha(sha, perPage, labels);
        if (commits.length) {
            checkpoint_1.checkpoint(`found ${commits.length} commits since ${sha}`, checkpoint_1.CheckpointType.Success);
        }
        else {
            checkpoint_1.checkpoint(`no commits found since ${sha}`, checkpoint_1.CheckpointType.Failure);
        }
        return commits;
    }
    gitHubInstance() {
        const [owner, repo] = parseGithubRepoUrl(this.repoUrl);
        return new github_1.GitHub({
            token: this.token,
            owner,
            repo,
            apiUrl: this.apiUrl,
            proxyKey: this.proxyKey,
        });
    }
    async openPR(sha, changelogEntry, updates, version) {
        const title = `chore: release ${version}`;
        const body = `:robot: I have created a release \\*beep\\* \\*boop\\* \n---\n${changelogEntry}\n\nThis PR was generated with [Release Please](https://github.com/googleapis/release-please).`;
        const pr = await this.gh.openPR({
            branch: `release-v${version}`,
            version,
            sha,
            updates,
            title,
            body,
            labels: this.labels,
        });
        // a return of -1 indicates that PR was not updated.
        if (pr > 0) {
            await this.gh.addLabels(pr, this.labels);
            await this.closeStaleReleasePRs(pr);
        }
    }
}
exports.ReleasePR = ReleasePR;
function changelogEmpty(changelogEntry) {
    return changelogEntry.split('\n').length === 1;
}
function updatePHPChangelogEntry(pkgKey, changelogEntry, entryUpdate) {
    {
        // Remove the first line of the entry, in favor of <summary>.
        // This also allows us to use the same regex for extracting release
        // notes (since the string "## v0.0.0" doesn't show up multiple times).
        const entryUpdateSplit = entryUpdate.split(/\r?\n/);
        entryUpdateSplit.shift();
        entryUpdate = entryUpdateSplit.join('\n');
    }
    return `${changelogEntry}

<details><summary>${pkgKey}</summary>

${entryUpdate}

</details>`;
}
//# sourceMappingURL=release-pr.js.map